/*

==== recursion(Xml code) ====

Recursion limiter. Many library header files include files like stdio.h
and internal/private header files, which
contain definitions which you probarly don't want to define or redefine.
On the other hand, the header file may not be possible to parse
unless these files are included. Recursion limit solves the problem. The header file
is parsed with all includes processed as normal. Afterwards, definitions in
included files can be marked as hidden with the hide attribute set to 1.

Looks for #pragma directives that define recursion level. If any
are found, will limit depth of #include recursion to the given level. Example:

#pragma JSEXT recursion 1
#include <stdio.h>

Will limit definitions to those which occur in stdio.h and excludes
those that are defined in files included by stdio.h.

#pragma JSEXT recursion 0

Limits definitions to those given in the code string itself.

#pragma JSEXT recursion off

Turns off recursion limit (default).

The recursion level is counted from the level of the
code string, even if the pragma directive occurs at a lower level. The recursion
limit may be changed by successive pragma directives.

Returns pruned xml tree.

*/

C.recursion=function(code) {
//	var ret=new XMLList;
	var stack=[];
	var tu; // translation unit
        var level=undefined;
	for each(tu in code.*) {
		var match;
		if (tu.name()=="pragma" &&  (match=tu.match(/JSEXT[ \t]+recursion[ \t]+([0-9]+)?/))) {
			if (match[1]==undefined) level=undefined;
			else level=Number(match[1])+1;
		}
		if (tu.name()=="line") {
			if (tu.@line==1) {
				stack[stack.length]=tu.@file;
			} else {
				stack.length--;
			}
//			print(stack,"\n");
		}
		if (level!=undefined && stack.length>level) tu.parent().replace(tu.childIndex(),undefined);
//		if (level==undefined || stack.length<=level) ret+=tu;
	}
//	return <C>{ret}</C>;
}

/*
	Calls splitdeclarator to split declarators
*/


C.splitdeclarators=function(code) {
	var decl;
	for each(decl in code..d) {
		C.splitdeclarator(decl);
	}
}

/*
	Separates declaration specifiers from declarators and
	changes a declaration of multiple variables into multiple
	declarations of single variables
*/

C.splitdeclarator=function(decl) {
	// Find declarators
	var declarators=new XMLList;
	var i,j;

	for (i=decl.*.length(); --i;) {
		switch(String(decl.*[i].name())) {
			case 'ptr':
			case 'ix':
			case 'p':
			case 'bitfield':
			case 'fd':
			case 'id':
				declarators+=decl.*[i];
				break;
			default:
				i=1;
		}
	}

	if (declarators.length()<=1) return;

	var declarator;

	var declaration_specifiers=new XMLList;
	for (i=0; i<decl.*.length()-declarators.length(); i++) {
		declaration_specifiers+=decl.*[i];
	}


	for (j=0; j<declarators.length(); j++) {
		var declarator=declarators[j];

 		var subdecl=new XMLList;
		subdecl+=declaration_specifiers;
		subdecl+=declarator;
		if (j==declarators.length()-1)
			decl.parent().replace(decl.childIndex(),<d>{subdecl}</d>);
		else
			decl.parent().insertChildAfter(decl,<d>{subdecl}</d>);
	}
}

C.getident=function(decl) {
	var i=decl.*.length();
	i--;
	switch(String(decl.*[i].name())) {
		case 'ptr':
		case 'ix':
		case 'p':
		case 'bitfield':
		case 'fd':
			return decl.*[i]..id[0];
			break;
		case 'id':
			return decl.*[i];
			break;
	}
}



