/*
	===help([symbol="help"], [format], [include_path=global.include_path])===

	Prints help about a function or object
*/

function help(symbol, format, include_path) {
	if (symbol==undefined) symbol="help";
	var text=help.find(symbol, include_path);

	switch(format) {
	default:
		if (text==undefined) {
			print("No help defined for ",symbol,"\n");
		} else {
			print(text);
			if (text.subtopics.length) {
				print("\nSubtopics:\n\n", help.indent_subtopics(text.subtopics), "\n");
			}
		}
	}
}

help.indent_subtopics=function(subtopics) {
	var i;
	var hierarchy={};
	for (i in subtopics) {
		var hpos=hierarchy;
		var subtopic=subtopics[i];
		var parts=subtopic.split(".");
		var j;
		for (j in parts) {
			if (!(parts[j] in hpos)) hpos[parts[j]]={};
			hpos=hpos[parts[j]];
		}
	}
	function indent(obj, indenttext) {
		var ret="";
		var i;
		var sorter=[];
		for (i in obj) {
			sorter.push(i);
		}
		sorter.sort();
		for (i=0; i<sorter.length; i++) {
			ret+=indenttext+sorter[i]+"\n"+indent(obj[sorter[i]],indenttext+"\t");
		}
		return ret;
	}

	return indent(hierarchy,"");
}

help.find=function(symbol, include_path) {
	if (include_path==undefined) include_path=global.include_path;
	var dotpos=symbol.indexOf(".");
	if (dotpos==-1) {
		var filename=symbol;
	} else {
		var filename=symbol.substr(0,dotpos);
	}

	function resolve(filename) {
		var full_filename;
		var i;
		for (i in include_path) {
			full_filename=include_path[i]+"/"+filename+".js";
			if (_.access(full_filename,_.R_OK)==0) {
				return full_filename;
			}
			full_filename=include_path[i]+"/"+filename+"/"+filename+".js";
			if (_.access(full_filename,_.R_OK)==0) {
				return full_filename;
			}
		}
	}

	var full_filename=resolve(filename);
	if (full_filename==undefined) return;

	var content="\n"+read(full_filename);
	var defpos=content.indexOf("\n"+symbol+"=");
	if (defpos==-1) {
		defpos=content.indexOf("\nfunction "+symbol);
	}

	var commstart=-1;
	var commstop=-1;

	if (defpos!=-1) {
		commstart=content.lastIndexOf("/*",defpos);
		commstop=content.lastIndexOf("*/",defpos);
	}

	var subtopics={};
	defpos=0;
	var lastcomstart=-1;
	for (;;) {
		defpos=content.indexOf("\n"+symbol+".",defpos)
		if (defpos==-1) break;
		var subtopic=content.substr(defpos+2+symbol.length,128);
		subtopic=subtopic.match(/[a-zA-Z\.0-9_]*/)[0];
		var localcomstart=content.lastIndexOf("/*",defpos);
		var localcomstop=content.lastIndexOf("*/",defpos);
		if (localcomstart!=-1 &&
			localcomstart!=lastcomstart &&
			localcomstart!=commstart &&
			localcomstop!=localcomstart+2) {
			subtopics[subtopic]=true;
		}
		defpos++;
		lastcomstart=localcomstart;
	}
	var st=[];
	for (var i in subtopics) st.push(symbol+"."+i);

	if (commstart!=-1 && commstop!=-1 && commstart<commstop) {
		var ret=new String(content.substr(commstart+2,commstop-commstart-2));
	} else {
		var ret=new String;
	}
	ret.filename=full_filename;
	ret.pos=commstart+2;
	ret.len=commstop-commstart-2;
	ret.subtopics=st;

	return ret;
}
