/*

===interactive()===

Uses the GNU readline/history library to read commands into jsext. Usually included by the default systemwide initialization file when no file is specified to jsext.

====Files====

;~/.jsext_history
:Used to store previous commands entered in an interactive jsext session. Updated only on a normal exit from jsext.

====See also====

[[jsext]], [[jsextrc]]

*/

module.include("jsapi");

function interactive() {
	environment.JSEXT_INTERACTIVE="yes";

	var normalprompt="jsext> ";
	var contprompt="> ";
	var cmdbuf="";

	if (global._WINDOWS==undefined && !_.isatty(0)) {
		var cons=stdin;
		cons.Options={};
	} else {
		var cons=console.open({
			prompt: normalprompt,
			histfile: environment.HOME+"/.jsext_history",
			ClassName: "jsext.interactive",
			WindowName: "jsext",
			Height: 600,
			Menu: [
		{
			STRING: "&File",
			SubMenu: [
				{
					STRING: "&Run...",
					OnCOMMAND: function() {
						var fn=windows.GetOpenFileName();
						if (fn.File!="") {
							var scr=read(fn.File);
							_.JS_EvaluateScript(cx, globptr, scr, scr.length, fn.File, 1, [0]);
						}
					}
				},
				{
					STRING: "E&xit",
					OnCOMMAND: function() {
						_.PostQuitMessage(0);
					}
				}
			]
		},
		{
			STRING: "F&ormat",
			SubMenu: [
				{
					STRING: "&Font...",
					OnCOMMAND: function() {
						var lf=windows.ChooseFont(cons.lf);
						if (lf) {
							cons.lf=lf;
							cons.font=windows.CreateFont(lf);
							cons.Edit.SETFONT(cons.font);
						}
					}
				}
			]
		}
	]
		});
	}

	var ret=InitContext();
	global.cx=ret[0];
	var glob=ret[1];
	glob.stdout=cons;
	global.globptr=_.JS_GetGlobalObject(cx);

	for(;;) {
		var line=cons.readline();
		if (cons.eof()) break;
		cmdbuf+=line;
		if (_.JS_BufferIsCompilableUnit(cx, globptr, cmdbuf, cmdbuf.length)) {
			var rval=[0];
			_.JS_EvaluateScript(cx, globptr, cmdbuf, cmdbuf.length, null, 1, rval);
			cmdbuf="";
			cons.Options.prompt=normalprompt;
		} else {
			cons.Options.prompt=contprompt;
		}
	}

	cons.close();
	return 0;
}
